from __future__ import nested_scopes

import QUB
from QUBB import *
from Buttons import *
from qubtree import Node
from ModelBuilder import ModelTree
from math import *
from sys import *

try:
    from datetime import datetime
except:
    pass

import win32com.client
from pythoncom import CoInitialize, CoUninitialize

#----- Log results to diff_sim.txt, df_time.txt
def WriteFileOut(fname, lst, cMode):
    f = open(fname, cMode)
    for i, x in enumerate(lst):
        if i:
            f.write('\t')
        f.write(str(x))
    f.write('\n')
    f.close()

#----- Write current data and time to file 
def WriteDateTime(WriteFunc):
    try:
        WriteFunc([datetime.now().strftime("%d %b %Y %H:%M:%S")])
    except:
        pass

OverWriteSimOut3 = lambda lst: WriteFileOut('diff_sim.txt', lst,'w')
WriteSimOut3 = lambda lst: WriteFileOut('diff_sim.txt', lst,'a')
OverWriteTimeOut3 = lambda lst: WriteFileOut('diff_Time.txt', lst,'w')
WriteTimeOut3 = lambda lst: WriteFileOut('diff_Time.txt', lst,'a')

#----- Global data                 
aEStep = []
aPhiStep = []
cPrevModel = ''

#---------------------------------------------------------------------------------------
#----- combine any # of lists of equal length as new list ( L0[0],L1[0], L0[1], L1[1], ... ) 
def interleave(*lists):
    out = []
    for i in range(len(lists[0])):
        for l in range(len(lists)):
            out.append( lists[l][i] )
    return out

def interleave3(*lists):
    out = []
    for i in range(len(lists[0])):
        for l in range(len(lists)):
            out.append( lists[l][i][3] )
    return out

#----- convert supplied 'CCOO' format string to (0,0,1,1) list.
def listStates(states):
    lst = []
    for c in states:
        if c == 'c':
            lst.append(0)
        elif c == 'o':
            lst.append(1)
    return lst

#----- Present dialogue to enter { E Step 1, Phi Step 1, ... , E Phi Step n-1, Phi Step n-1 }
#----- Results are stored to global lists aEStep, aPhiStep
def PromptSteps3(n):
    #-- append to prompts array if more states than last time.
    shortfall = (n-1) - len(aEStep)
    while len(aEStep) < n-1 : 
        aEStep.append( [0.0,0.0,0.0,0.0,'0.0'])
        aPhiStep.append( [0.5,0.5,0.0,0.5,'0.5'])

    #-- Build and execute dialogue for values/ranges
    fields = []
    for i in range(n-1):
        fields.append(DialogText('E%d'%i, 'E step %d'%(i+1), aEStep[i][4] ))
        fields.append(DialogText('phi%d'%i, 'phi step %d'%(i+1), aPhiStep[i][4] ))
    vals = QUB.Dialog('Enter steps ( value or min;max;step )', ['OK'], fields)

    #-- Build lists of parameter ranges [min,max,step,initial]
    for i in range(n-1):
        c = vals['E%d'%i]
        iSCPos0 = c.find(';')
        iSCPos1 = c.rfind(';')
        try:
            if iSCPos0<=0 or iSCPos1==-1 or iSCPos0==iSCPos1 :
                aEStep[i] =  [float(c), float(c), 0.0, float(c),c]
            else:
                aEStep[i] =  [float(c[:iSCPos0]), float(c[iSCPos0+1:iSCPos1]), float(c[iSCPos1+1:]),float(c[:iSCPos0]),c]
                #-- Ensure step variable is + for increments or - for decrements
                if (aEStep[i][0]<aEStep[i][1]) != (aEStep[i][2]>0) :
                    aEStep[i][2] = -1 * aEStep[i][2]
        except:
            aEStep[i] = [0.0,0.0,0.0,0.0,aEStep[i][4]] 

        c = vals['phi%d'%i]
        iSCPos0 = c.find(';')
        iSCPos1 = c.rfind(';')
        try:
            if iSCPos0<=0 or iSCPos1== -1 or iSCPos0==iSCPos1 :
                aPhiStep[i] = [float(c), float(c), 0.0, float(c),c]
            else:
                aPhiStep[i] = [float(c[:iSCPos0]), float(c[iSCPos0+1:iSCPos1]), float(c[iSCPos1+1:]),float(c[:iSCPos0]),c]
                #-- Ensure step variable is + for increments or - for decrements
                if (aPhiStep[i][0]<aPhiStep[i][1]) != (aPhiStep[i][2]>0) :
                    aPhiStep[i][2] = -1 * aPhiStep[i][2]
        except:
            aPhiStep[i] = [0.0,0.0,0.0,0.0,aPhiStep[i][4]] 
    
#----- Add states  specified in string to model 
def DiffusionModel(props):
    #----- convert CCOO string to states list 
    states = listStates(props['states'])
    n = len(states)
    k = props['k']

    #----- get current model 
    modelObj = QUB.ModelView.currModel
    if not modelObj:
        QUB.Report('no model')
        return


    #----- Add rightmost (n-1) state
    modelTree = ModelTree()
    modelTree.node['ChannelCount'].data = props['SimChannels']

    state = modelTree.addState( states[n-1] )
    state['x'].data = 85.0
    state['y'].data = 50.0
    
    #----- add states right to left, each connected to prior state 
    for i in range(1,n):
        step_ix = (n-1) - i
        state = modelTree.addStateFrom(i-1, states[step_ix])
        state['x'].data = 85.0 - 70 * i / (n-1)
        state['y'].data = 50.0
        rates = modelTree.rates[i-1]['k0'].data
        rates[0] = k * pow( exp(aEStep[step_ix][3]), aPhiStep[step_ix][3] - 1 )
        rates[1] = k * pow( exp(aEStep[step_ix][3]), aPhiStep[step_ix][3] )

    #----- Update model 
    modelObj.setAsNode( modelTree.node )

def DiffusionRates_Solve3(props, modelObj):
    #----- Build model 
    DiffusionModel(props)

    states = listStates(props['states'])
    n = len(states)

    if props['Mode'] == 'solve':    
        #----- Simulate some data 
        Run( 'Simulation', 'Sim' )

        k12, k21 = props['k12'], props['k21']

        modelTree = ModelTree()
        modelTree.node['ChannelCount'].data = props['MilChannels']
        state = modelTree.addState(0)
        state['x'].data = 30.0
        state['y'].data = 50.0
        state = modelTree.addStateFrom(0, 1)
        state['x'].data = 70.0
        state['y'].data = 50.0
        rates = modelTree.rates[0]['k0'].data
        rates[0] = k12 * 0.1
        rates[1] = k21 * 0.1

        modelObj.setAsNode( modelTree.node )
        milPrefs = Node()
        milPrefs['Mode'].data = 'optimize'
        Run( 'Modeling', 'MIL', milPrefs )

        rates = modelObj.getAsNode()['Rates']['Rate']['k0'].data
        L = (rates[0] / rates[1])
        try:
            fee = log(rates[0] / k12) / log((rates[0] / k12) * (k21 / rates[1]))
            if (props['TextOut'] == 'Append Full' or props['TextOut'] == 'Overwrite Full') : 
                row = [props['states'], props['k'], k12, k21, rates[0], rates[1], L, fee]
                row = row + interleave3(aEStep[:n-1], aPhiStep[:n-1])
            else :
                row = [log10(L), log10(rates[0]), log10(rates[1])]
            err = QUB.ResultsView.currResults.node.find('ErrorCode').data
            if len(err) and err[0]:
                row.append('Error: %d' % err[0])
                
            WriteSimOut3(row)
        except:
            pass

    else: # 'time constants'
        nBlack, nRed = 0, 0
        for state in states:
            if state == 0:
                nBlack = nBlack + 1
            elif state == 1:
                nRed = nRed + 1

        evalOnly = Node()
        evalOnly['Mode'].data = 'evaluate'
        Run('Modeling', 'MIL', evalOnly)
        if (props['TextOut'] == 'Append Full' or props['TextOut'] == 'Overwrite Full') : 
            row = [props['states'], props['k']]
        else:
            row=[]
            
        tc = QUB.ResultsView.currResults.node.find('DataSet').find('HistogramGroup').find('TimeConstants')
        row = row + [tc['Tau'].data[nBlack-1], tc['Amp'].data[nBlack-1]]
        alltc = [' '] + interleave(tc['Tau'].data, tc['Amp'].data)
        tc = tc.next('TimeConstants')
        row = row + [tc['Tau'].data[nRed-1], tc['Amp'].data[nRed-1]]

        if (props['TextOut'] == 'Append Full' or props['TextOut'] == 'Overwrite Full') : 
            alltc = alltc + [' '] + interleave(tc['Tau'].data, tc['Amp'].data)
            row = row + interleave3(aEStep[:n-1], aPhiStep[:n-1]) + alltc
        WriteTimeOut3(row)

#----- Go to the next iteration.  Return True if iteration succeeded.
# increment the rightmost range and backtrack on rollover.  
def Increment3(i):
    ret=True

    if (i%2) == 1 :
        rng = aPhiStep[(i-1)/2]
    else:
        rng = aEStep[i/2]

    if (rng[0]==rng[1]) or ((rng[0]<rng[1]) and (rng[3]>=rng[1])) or ((rng[0]>rng[1]) and (rng[3]<=rng[1])):
        rng[3]=rng[0]
        if i==0 :
            ret=False
        else:
            ret = Increment3(i-1)
    else:
        rng[3]=rng[3]+rng[2]
            
    return ret

#----- Main function - diffusion with all rates able to be set as ranges
def Diffusion3(props):
    global cPrevModel, aEStep, aPhiStep
    
    #-- Check for current model 
    modelObj = QUB.ModelView.currModel
    if not modelObj:
        QUB.Report('no model')
        return

    #-- convert 
    states = listStates(props['states'])
    n = len(states)

    #----- If model changes , clear previous states values
    if cPrevModel != props['states'] :
        aEStep = []
        aPhiStep = []
    cPrevModel=props['states']
    
    #-- Create a list of lists : {Min,Max,Step,Current=min}
    PromptSteps3(n)

    #----- Write file header 
    if (props['Mode'] == 'create model') : 
        DiffusionModel(props)
    else : # 'solve' , 'time constants'
        if (props['Mode'] == 'solve') : 
            if (props['TextOut'] == 'Append Full') or (props['TextOut'] == 'Append Terse'): 
                WriteSimOut3([])
                WriteDateTime(WriteSimOut3)
            else:
                OverWriteSimOut3([])
            if (props['TextOut'] == 'Append Full') or (props['TextOut'] == 'Overwrite Full'): 
                WriteSimOut3(['states', 'k', 'k12', 'k21', 'k+', 'k-', 'L', 'phi', 'steps (E, phi)'])
            else :
                WriteSimOut3(['Log(L)','Log(k+)', 'Log(k-)'])
            WriteOut = WriteSimOut3
        else:
            if (props['TextOut'] == 'Append Full') or (props['TextOut'] == 'Append Terse'): 
                WriteTimeOut3([])
                WriteDateTime(WriteTimeOut3)
            else:
                OverWriteTimeOut3([])
            if (props['TextOut'] == 'Append Full') or (props['TextOut'] == 'Overwrite Full'): 
                WriteTimeOut3(['states', 'k', 'Tau black', 'Amp black', 'Tau red', 'Amp red', 'steps (E, phi)'])
            else :
                WriteTimeOut3(['Tau black', 'Amp black', 'Tau red', 'Amp red'])
            WriteOut = WriteTimeOut3

        #-- Loop until all are exhausted.
        lCont = True
        iter = 0 
        while lCont and (iter<500) :     
            DiffusionRates_Solve3(props, modelObj)
            lCont=Increment3(2*(n-1)-1)
            iter = iter + 1
        if (iter>=500) :
            WriteOut(['Too many iterations.  500 is maximum.'])
    
        WriteOut([])

#----- Set State list validation 
def acceptStates(s):
    s = s.lower()
    for c in s:
        if (c != 'c') and (c != 'o'):
            raise FieldAcceptException(s, "must be a string of C and O")
    if len(s) < 2:
        raise FieldAcceptException(s, "must have at least two states")

    return s

#----- Add button to the QUB Modelling button group
AddTaskButton( 'Dif v. 3', # button text
               Diffusion3,   # function
               'Dif3',       # short text for small buttons
               0,            # show settings dialog on left-click?
               1,            # show settings dialog on right-click?
               "Ask Tony Auerbach",   # tool-tip
               "Ask Tony Auerbach",   # description in "Edit Buttons"
               "Diffusion rates properties:", # caption for the settings dialog
               [TextField("states", "States:", "CCCOO", acceptStates),
                ChoiceField("Mode", "Mode:", ['create model', 'solve', 'time constants'], 'solve'),
                ChoiceField("TextOut", "Text Out:", ['Append Full', 'Append Terse', 'Overwrite Full', 'Overwrite Terse'], 'Append Full'),
                TextField("SimChannels", "Sim Channels:", "1", acceptIntGreaterThan(0)),
                TextField("MilChannels", "Mil Channels:", "1", acceptIntGreaterThan(0)),
                TextField("k", "k:", "10000", acceptFloat),
                TextField("k12", "k12:", "10000", acceptFloat),
                TextField("k21", "k21:", "10000", acceptFloat)],
               'Modeling' )
